#!/usr/bin/python
# Copyright (C) Patrick Brady, Brian Moe, Branson Stephens (2015)
#
# This file is part of lvalert
#
# lvalert is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lvalert.  If not, see <http://www.gnu.org/licenses/>.

from netrc import netrc, NetrcParseError
import stat
import sys
import os
if os.name == 'posix':
    import pwd
from six.moves.urllib.parse import urlparse
from six import StringIO
from optparse import *
from subprocess import Popen,PIPE

# Note that some of these utilities are not available unless glue is installed.

import warnings
import json

try:
    from glue.ligolw import ligolw
    from glue.ligolw import table
    from glue.ligolw.table import Table
    from glue.ligolw import lsctables
    from glue.ligolw import utils
except ImportError:
    Table = object
else:
    class LIGOLWContentHandler(ligolw.LIGOLWContentHandler):
        pass
    lsctables.use_in(LIGOLWContentHandler)

##############################################################################
#
#          definition of the LVAlertTable
#
##############################################################################

class LVAlertTable(Table):
  """
  for reference, file is written as
  file: //host/path_to_file/file
  uid: the unique id assigned by gracedb
  temp_data_loc: current location (just the directory)
                 of the ouput of the pipeline (this is VOLATILE)
  """
  tableName = "lvalert:table"
  validcolumns = {
    "file": "lstring",
    "uid": "lstring",
    "temp_data_loc": "lstring",
    "alert_type": "lstring",
    "description": "lstring",
    }
    
class LVAlertRow(object):
  __slots__ = list(LVAlertTable.validcolumns.keys())
  
LVAlertTable.RowType = LVAlertRow

##############################################################################
#
#          useful utilities
#
##############################################################################

def parse_file_url(file_url):
  """
  simple function to parse the file urls of the form:
  file://host_name/path_to_file/file
  where path_to_file is assumed to be the /private subdir of a gracedb entry
  returns:
  host: host_name in the above example
  full_path: path_to_file/file in the above example
  general_dir: the /general subdir of the gracedb entry (where data not
  produced by the event supplier should be placed)
  """
  parsed = urlparse(file_url)
  host = parsed[1]
  path, fname = os.path.split(parsed[2])
  
  return host, path, fname

def get_LVAdata_from_stdin(std_in, as_dict=False):
  """
  this function takes an LVAlertTable *OR* a JSON-serialized dictionary from 
  sys.stdin and it returns:

  a full dictionary of the LVAlertTable values, or:

  file: the filename (if any) associated with the alert
  uid: the gracedb unique id associated with the event in the LVAlertTable
  data_loc: a URL for the payload file
  """
  warnings.warn("get_LVAdata_from_stdin is deprecated. Use Python's json module: json.loads(stdin.read())")
  content = std_in.read()
  # Try interpreting it as JSON first.
  try:
    out_dict = json.loads(content)
    file = out_dict['file']
    uid  = out_dict['uid']
    data_loc = out_dict['data_loc']
    description = out_dict['description']
    alert_type = out_dict['alert_type']
  except Exception as e:
    # We don't have a file object anymore, because we .read() it.
    # Instead, we want this to load a blob of text. 
    f = StringIO(content)
    doc = utils.load_fileobj(f, contenthandler = LIGOLWContentHandler)[0]
    lvatable = table.get_table(doc, LVAlertTable.tableName)
    file = lvatable[0].file
    uid = lvatable[0].uid
    data_loc = lvatable[0].temp_data_loc
    description = lvatable[0].description
    alert_type = lvatable[0].alert_type
  if as_dict:
    return {
      "file"        : file,
      "uid"         : uid,
      "data_loc"    : data_loc,
      "description" : description,
      "alert_type"  : alert_type,
    }
  return file, uid, data_loc

def get_LVAdata_from_file(filename, as_dict=False):
  """
  this function takes the name of an xml file containing a single LVAlertTable
  and it returns:
  host: the machine the payload file was created on
  full_path: the full path to (and including) the payload file
  general_dir: the directory in gracedb that the output of your code should
               be written to
  uid: the gracedb unique id associated with the event in the LVAlertTable
  """
  doc = utils.load_filename(filename, contenthandler = LIGOLWContentHandler)
  lvatable = table.get_table(doc, LVAlertTable.tableName)
  file = lvatable[0].file
  uid = lvatable[0].uid
  data_loc = lvatable[0].temp_data_loc

  if as_dict:
    return {
      "file" : lvatable[0].file,
      "uid" : lvatable[0].uid,
      "data_loc" : lvatable[0].temp_data_loc,
      "description" : lvatable[0].description,
      "alert_type" : lvatable[0].alert_type,
    }

  return file, uid, data_loc  

def make_LVAlertTable(file_url, uid, data_loc, alert_type="new", desc=""):
  """
  create xml doc which contains an LVAlert Table
  with submission file file_loc and  data located at data_loc
  """
  xmldoc = ligolw.Document()
  xmldoc.appendChild(ligolw.LIGO_LW())
  lvalerttable = lsctables.New(LVAlertTable)
  row = lvalerttable.RowType()
  row.file = file_url
  row.uid = uid
  row.temp_data_loc = data_loc
  row.alert_type = alert_type
  row.description = desc
  lvalerttable.append(row)
  xmldoc.childNodes[0].appendChild(lvalerttable)

  return xmldoc


#the following is meant as a template for small jobs
#notes:
#   * we only use the vanilla universe which is appropriate for python
#    jobs and things not condor-compiled
#   * it only works for single-process jobs; anything more complicated will
#    require a dag
condor_sub_template = \
                    """
                    universe = vanilla
                    executable = macroexecutible
                    arguments = macroargs
                    log = macrolog
                    error = macroerr
                    output = macroout
                    getenv = True
                    notification = never
                    queue
                    """
def write_condor_sub(executible, args, logdir, uid):
  """
  write a simple condor submission file
  uid: unique id used in naming the files (to avoid conflicts)
  executible: the name of the executible file
  args: a list of arguments to executible
  logdir: directory to keep log files
  returns the name of the file
  """
  subfile = condor_sub_template.replace('macroexecutible', executible)\
            .replace('macroargs', args)\
            .replace('macrolog', os.path.join(logdir,str(uid)+'.log'))\
            .replace('macroerr', os.path.join(logdir,str(uid)+'.err'))\
            .replace('macroout', os.path.join(logdir,str(uid)+'.out'))
  fname = str(uid) + '.sub'
  f = open(fname,'w')
  f.write(subfile)
  f.close()

  return fname

def submit_condor_job(subfile):
  """
  submit the subfile to condor
  returns the process id
  """
  p = Popen(["condor_submit "+subfile], shell=True).pid
  
  return p

class safe_netrc(netrc):
    """The netrc.netrc class from the Python standard library applies access
    safety checks (requiring that the netrc file is readable only by the
    current user, and not by group members or other users) only if using the
    netrc file in the default location (~/.netrc). This subclass applies the
    same access safety checks regardless of the path to the netrc file."""

    def _parse(self, file, fp, *args, **kwargs):
        # Copied and adapted from netrc.py from Python 2.7
        if os.name == 'posix':
            prop = os.fstat(fp.fileno())
            if prop.st_uid != os.getuid():
                try:
                    fowner = pwd.getpwuid(prop.st_uid)[0]
                except KeyError:
                    fowner = 'uid %s' % prop.st_uid
                try:
                    user = pwd.getpwuid(os.getuid())[0]
                except KeyError:
                    user = 'uid %s' % os.getuid()
                raise NetrcParseError(
                    ("~/.netrc file owner (%s) does not match"
                     " current user (%s)") % (fowner, user),
                    file)
            if (prop.st_mode & (stat.S_IRWXG | stat.S_IRWXO)):
                raise NetrcParseError(
                   "~/.netrc access too permissive: access"
                   " permissions must restrict access to only"
                   " the owner", file)
        return netrc._parse(self, file, fp, *args, **kwargs)



# Copyright (C) Patrick Brady, Brian Moe, Branson Stephens (2015)
# based on work by Jacek Konieczny <jajcus@jajcus.net> 2003-2004
#
# This file is part of lvalert
#
# lvalert is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lvalert.  If not, see <http://www.gnu.org/licenses/>.

"""Handling of XMPP stanzas.

Normative reference:
  - `RFC 3920 <http://www.ietf.org/rfc/rfc3920.txt>`__
"""

__revision__="$Id$"
__docformat__="restructuredtext en"

## Needed because openfire has a minor bug associated with results
## being ignored in the handshake stage. There is a one line fix
## in this overloaded class.
import libxml2
import logging
import threading

import pyxmpp.stanzaprocessor
from pyxmpp.expdict import ExpiringDictionary
from pyxmpp.exceptions import ProtocolError, BadRequestProtocolError, FeatureNotImplementedProtocolError
from pyxmpp.stanza import Stanza
from pyxmpp.stanzaprocessor import StanzaProcessor as LVAlertSP

class LVAlertStanzaProcessor(LVAlertSP):
    """Universal stanza handler/router class.

    Provides facilities to set up custom handlers for various types of stanzas.

    :Ivariables:
        - `lock`: lock object used to synchronize access to the
          `StanzaProcessor` object.
        - `me`: local JID.
        - `peer`: remote stream endpoint JID.
        - `process_all_stanzas`: when `True` then all stanzas received are
          considered local.
        - `initiator`: `True` if local stream endpoint is the initiating entity.
    """
    def __init__(self):
        """Initialize a `StanzaProcessor` object."""
        self.me=None
        self.peer=None
        self.initiator=None
        self.peer_authenticated=False
        self.process_all_stanzas=True
        self._iq_response_handlers=ExpiringDictionary()
        self._iq_get_handlers={}
        self._iq_set_handlers={}
        self._message_handlers=[]
        self._presence_handlers=[]
        self.__logger=logging.getLogger("pyxmpp.Stream")
        self.lock=threading.RLock()

    def process_response(self, response):
        """Examines out the response returned by a stanza handler and sends all
        stanzas provided.

        :Returns:
           - `True`: if `response` is `Stanza`, iterable or `True` (meaning the stanza was processed).
           - `False`: when `response` is `False` or `None`
        :returntype: `bool`
        """

        if response is None or response is False:
            return False

        if isinstance(response, Stanza):
            self.send(response)
            return True

        try:
            response = iter(response)
        except TypeError:
            return bool(response)

        for stanza in response:
            if isinstance(stanza, Stanza):
                self.send(stanza)
        return True

    def process_iq(self, stanza):
        """Process IQ stanza received.

        :Parameters:
            - `stanza`: the stanza received

        If a matching handler is available pass the stanza to it.
        Otherwise ignore it if it is "error" or "result" stanza
        or return "feature-not-implemented" error."""

        sid=stanza.get_id()
        fr=stanza.get_from()

        typ=stanza.get_type()
        if typ in ("result","error"):
            if fr:
                ufr=fr.as_unicode()
            else:
                ufr=None
            if (sid,ufr) in self._iq_response_handlers:
                key=(sid,ufr)
            elif ( (fr==self.peer or fr==self.me)
                    and (sid,None) in self._iq_response_handlers):
                key=(sid,None)
            else:
                return False
            res_handler, err_handler = self._iq_response_handlers[key]
            if stanza.get_type()=="result":
                response = res_handler(stanza)
            else:
                response = err_handler(stanza)
            del self._iq_response_handlers[key]
            self.process_response(response)
            return True

        q=stanza.get_query()
        if not q:
            raise BadRequestProtocolError("Stanza with no child element")
        el=q.name
        ns=q.ns().getContent()

        if typ=="get":
            if (el,ns) in self._iq_get_handlers:
                response = self._iq_get_handlers[(el,ns)](stanza)
                self.process_response(response)
                return True
            else:
                raise FeatureNotImplementedProtocolError("Not implemented")
        elif typ=="set":
            if (el,ns) in self._iq_set_handlers:
                response = self._iq_set_handlers[(el,ns)](stanza)
                self.process_response(response)
                return True
            else:
                raise FeatureNotImplementedProtocolError("Not implemented")
        else:
            raise BadRequestProtocolError("Unknown IQ stanza type")

    def __try_handlers(self,handler_list,typ,stanza):
        """ Search the handler list for handlers matching
        given stanza type and payload namespace. Run the
        handlers found ordering them by priority until
        the first one which returns `True`.

        :Parameters:
            - `handler_list`: list of available handlers
            - `typ`: stanza type (value of its "type" attribute)
            - `stanza`: the stanza to handle

        :return: result of the last handler or `False` if no
            handler was found."""
        namespaces=[]
        if stanza.xmlnode.children:
            c=stanza.xmlnode.children
            while c:
                try:
                    ns=c.ns()
                except libxml2.treeError:
                    ns=None
                if ns is None:
                    c=c.next
                    continue
                ns_uri=ns.getContent()
                if ns_uri not in namespaces:
                    namespaces.append(ns_uri)
                c=c.next
        for handler_entry in handler_list:
            t=handler_entry[1]
            ns=handler_entry[2]
            handler=handler_entry[3]
            if t!=typ:
                continue
            if ns is not None and ns not in namespaces:
                continue
            response = handler(stanza)
            if self.process_response(response):
                return True
        return False

    def process_message(self,stanza):
        """Process message stanza.

        Pass it to a handler of the stanza's type and payload namespace.
        If no handler for the actual stanza type succeeds then hadlers
        for type "normal" are used.

        :Parameters:
            - `stanza`: message stanza to be handled
        """

        if not self.initiator and not self.peer_authenticated:
            self.__logger.debug("Ignoring message - peer not authenticated yet")
            return True

        typ=stanza.get_type()
        if self.__try_handlers(self._message_handlers,typ,stanza):
            return True
        if typ!="error":
            return self.__try_handlers(self._message_handlers,"normal",stanza)
        return False

    def process_presence(self,stanza):
        """Process presence stanza.

        Pass it to a handler of the stanza's type and payload namespace.

        :Parameters:
            - `stanza`: presence stanza to be handled
        """

        if not self.initiator and not self.peer_authenticated:
            self.__logger.debug("Ignoring presence - peer not authenticated yet")
            return True

        typ=stanza.get_type()
        if not typ:
            typ="available"
        return self.__try_handlers(self._presence_handlers,typ,stanza)

    def route_stanza(self,stanza):
        """Process stanza not addressed to us.

        Return "recipient-unavailable" return if it is not
        "error" nor "result" stanza.

        This method should be overriden in derived classes if they
        are supposed to handle stanzas not addressed directly to local
        stream endpoint.

        :Parameters:
            - `stanza`: presence stanza to be processed
        """
        if stanza.get_type() not in ("error","result"):
            r = stanza.make_error_response("recipient-unavailable")
            self.send(r)
        return True

    def process_stanza(self,stanza):
        """Process stanza received from the stream.

        First "fix" the stanza with `self.fix_in_stanza()`,
        then pass it to `self.route_stanza()` if it is not directed
        to `self.me` and `self.process_all_stanzas` is not True. Otherwise
        stanza is passwd to `self.process_iq()`, `self.process_message()`
        or `self.process_presence()` appropriately.

        :Parameters:
            - `stanza`: the stanza received.

        :returns: `True` when stanza was handled
        """

        # PRB changes start
        self.fix_in_stanza(stanza)
        to=stanza.get_to()
        type=stanza.get_type()

        if not self.process_all_stanzas and to and to!=self.me and to.bare()!=self.me.bare() and type!="result":
        #if not self.process_all_stanzas and to and to!=self.me and to.bare()!=self.me.bare():
            return self.route_stanza(stanza)
        # PRB changes end

        try:
            if stanza.stanza_type=="iq":
                if self.process_iq(stanza):
                    return True
            elif stanza.stanza_type=="message":
                if self.process_message(stanza):
                    return True
            elif stanza.stanza_type=="presence":
                if self.process_presence(stanza):
                    return True
        except ProtocolError as e:
            typ = stanza.get_type()
            if typ != 'error' and (typ != 'result' or stanza.stanza_type != 'iq'):
                r = stanza.make_error_response(e.xmpp_name)
                self.send(r)
                e.log_reported()
            else:
                e.log_ignored()

        self.__logger.debug("Unhandled %r stanza: %r" % (stanza.stanza_type,stanza.serialize()))
        return False

    def check_to(self,to):
        """Check "to" attribute of received stream header.

        :return: `to` if it is equal to `self.me`, None otherwise.

        Should be overriden in derived classes which require other logic
        for handling that attribute."""
        if to!=self.me:
            return None
        return to

    def set_response_handlers(self,iq,res_handler,err_handler,timeout_handler=None,timeout=300):
        """Set response handler for an IQ "get" or "set" stanza.

        This should be called before the stanza is sent.

        :Parameters:
            - `iq`: an IQ stanza
            - `res_handler`: result handler for the stanza. Will be called
              when matching <iq type="result"/> is received. Its only
              argument will be the stanza received. The handler may return
              a stanza or list of stanzas which should be sent in response.
            - `err_handler`: error handler for the stanza. Will be called
              when matching <iq type="error"/> is received. Its only
              argument will be the stanza received. The handler may return
              a stanza or list of stanzas which should be sent in response
              but this feature should rather not be used (it is better not to
              respond to 'error' stanzas).
            - `timeout_handler`: timeout handler for the stanza. Will be called
              when no matching <iq type="result"/> or <iq type="error"/> is
              received in next `timeout` seconds. The handler should accept
              two arguments and ignore them.
            - `timeout`: timeout value for the stanza. After that time if no
              matching <iq type="result"/> nor <iq type="error"/> stanza is
              received, then timeout_handler (if given) will be called.
        """
        self.lock.acquire()
        try:
            self._set_response_handlers(iq,res_handler,err_handler,timeout_handler,timeout)
        finally:
            self.lock.release()

    def _set_response_handlers(self,iq,res_handler,err_handler,timeout_handler=None,timeout=300):
        """Same as `Stream.set_response_handlers` but assume `self.lock` is acquired."""
        self.fix_out_stanza(iq)
        to=iq.get_to()
        if to:
            to=to.as_unicode()
        if timeout_handler:
            self._iq_response_handlers.set_item((iq.get_id(),to),
                    (res_handler,err_handler),
                    timeout,timeout_handler)
        else:
            self._iq_response_handlers.set_item((iq.get_id(),to),
                    (res_handler,err_handler),timeout)

    def set_iq_get_handler(self,element,namespace,handler):
        """Set <iq type="get"/> handler.

        :Parameters:
            - `element`: payload element name
            - `namespace`: payload element namespace URI
            - `handler`: function to be called when a stanza
              with defined element is received. Its only argument
              will be the stanza received. The handler may return a stanza or
              list of stanzas which should be sent in response.

        Only one handler may be defined per one namespaced element.
        If a handler for the element was already set it will be lost
        after calling this method.
        """
        self.lock.acquire()
        try:
            self._iq_get_handlers[(element,namespace)]=handler
        finally:
            self.lock.release()

    def unset_iq_get_handler(self,element,namespace):
        """Remove <iq type="get"/> handler.

        :Parameters:
            - `element`: payload element name
            - `namespace`: payload element namespace URI
        """
        self.lock.acquire()
        try:
            if (element,namespace) in self._iq_get_handlers:
                del self._iq_get_handlers[(element,namespace)]
        finally:
            self.lock.release()

    def set_iq_set_handler(self,element,namespace,handler):
        """Set <iq type="set"/> handler.

        :Parameters:
            - `element`: payload element name
            - `namespace`: payload element namespace URI
            - `handler`: function to be called when a stanza
              with defined element is received. Its only argument
              will be the stanza received. The handler may return a stanza or
              list of stanzas which should be sent in response.


        Only one handler may be defined per one namespaced element.
        If a handler for the element was already set it will be lost
        after calling this method."""
        self.lock.acquire()
        try:
            self._iq_set_handlers[(element,namespace)]=handler
        finally:
            self.lock.release()

    def unset_iq_set_handler(self,element,namespace):
        """Remove <iq type="set"/> handler.

        :Parameters:
            - `element`: payload element name.
            - `namespace`: payload element namespace URI."""
        self.lock.acquire()
        try:
            if (element,namespace) in self._iq_set_handlers:
                del self._iq_set_handlers[(element,namespace)]
        finally:
            self.lock.release()

    def __add_handler(self,handler_list,typ,namespace,priority,handler):
        """Add a handler function to a prioritized handler list.

        :Parameters:
            - `handler_list`: a handler list.
            - `typ`: stanza type.
            - `namespace`: stanza payload namespace.
            - `priority`: handler priority. Must be >=0 and <=100. Handlers
              with lower priority list will be tried first."""
        if priority<0 or priority>100:
            raise ValueError("Bad handler priority (must be in 0:100)")
        handler_list.append((priority,typ,namespace,handler))
        handler_list.sort()

    def set_message_handler(self, typ, handler, namespace=None, priority=100):
        """Set a handler for <message/> stanzas.

        :Parameters:
            - `typ`: message type. `None` will be treated the same as "normal",
              and will be the default for unknown types (those that have no
              handler associated).
            - `namespace`: payload namespace. If `None` that message with any
              payload (or even with no payload) will match.
            - `priority`: priority value for the handler. Handlers with lower
              priority value are tried first.
            - `handler`: function to be called when a message stanza
              with defined type and payload namespace is received. Its only
              argument will be the stanza received. The handler may return a
              stanza or list of stanzas which should be sent in response.

        Multiple <message /> handlers with the same type/namespace/priority may
        be set. Order of calling handlers with the same priority is not defined.
        Handlers will be called in priority order until one of them returns True or
        any stanza(s) to send (even empty list will do).
        """
        self.lock.acquire()
        try:
            if not typ:
                typ=="normal"
            self.__add_handler(self._message_handlers,typ,namespace,priority,handler)
        finally:
            self.lock.release()

    def set_presence_handler(self,typ,handler,namespace=None,priority=100):
        """Set a handler for <presence/> stanzas.

        :Parameters:
            - `typ`: presence type. "available" will be treated the same as `None`.
            - `namespace`: payload namespace. If `None` that presence with any
              payload (or even with no payload) will match.
            - `priority`: priority value for the handler. Handlers with lower
              priority value are tried first.
            - `handler`: function to be called when a presence stanza
              with defined type and payload namespace is received. Its only
              argument will be the stanza received. The handler may return a
              stanza or list of stanzas which should be sent in response.

        Multiple <presence /> handlers with the same type/namespace/priority may
        be set. Order of calling handlers with the same priority is not defined.
        Handlers will be called in priority order until one of them returns
        True or any stanza(s) to send (even empty list will do).
        """
        self.lock.acquire()
        try:
            if not typ:
                typ="available"
            self.__add_handler(self._presence_handlers,typ,namespace,priority,handler)
        finally:
            self.lock.release()

    def fix_in_stanza(self,stanza):
        """Modify incoming stanza before processing it.

        This implementation does nothig. It should be overriden in derived
        classes if needed."""
        pass

    def fix_out_stanza(self,stanza):
        """Modify outgoing stanza before sending into the stream.

        This implementation does nothig. It should be overriden in derived
        classes if needed."""
        pass


    def send(self,stanza):
        """Send a stanza somwhere. This one does nothing. Should be overriden
        in derived classes.

        :Parameters:
            - `stanza`: the stanza to send.
        :Types:
            - `stanza`: `pyxmpp.stanza.Stanza`"""
        raise NotImplementedError("This method must be overriden in derived classes.""")

pyxmpp.stanzaprocessor.StanzaProcessor = LVAlertStanzaProcessor

# end of overloaded class

# vi: sts=4 et sw=4



from __future__ import print_function
import argparse
import logging
import sys

import sleekxmpp

from ligo.lvalert import LVAlertClient, DEFAULT_SERVER


def parser():
    parser = argparse.ArgumentParser(prog='lvalert')
    parser.add_argument('-l', '--log', help='Log level', default='error',
                        choices='critical error warning info debug'.split())
    parser.add_argument('-n', '--netrc',
                        help='netrc file (default: read from NETRC '
                        'environment variable or ~/.netrc)')
    parser.add_argument('-r', '--resource',
                        help='XMPP resource (default: random)')
    parser.add_argument('-s', '--server', default=DEFAULT_SERVER,
                        help='LVAlert server hostname')
    parser.add_argument('-u', '--username',
                        help='User name (default: look up in netrc file)')

    subparsers = parser.add_subparsers(dest='action', help='sub-command help')
    subparsers.required = True

    subparser = subparsers.add_parser(
        'listen', help='Listen for LVAlert messages and print them to stdout.')

    subparser = subparsers.add_parser(
        'subscriptions', help='List your subscriptions')

    subparser = subparsers.add_parser(
        'nodes', help='List available pubsub nodes')

    subparser = subparsers.add_parser(
        'subscribe', help='Subscribe to one or more nodes')
    subparser.add_argument(
        'node', nargs='+', help='a pubsub node (e.g. cbc_gstlal)')

    subparser = subparsers.add_parser(
        'unsubscribe', help='Unsubscribe from one or more nodes')
    subparser.add_argument(
        'node', nargs='+', help='a pubsub node (e.g. cbc_gstlal)')

    subparser = subparsers.add_parser(
        'send', help='publish contents of a file to a pubsub node')
    subparser.add_argument(
        'node', nargs='+', help='a pubsub node (e.g. cbc_gstlal)')
    subparser.add_argument(
        'eventfile', nargs='+', help='name of the file with the event to send',
                type=argparse.FileType('rb'))
    return parser


def show(node, payload):
    print('Got item for node:', node)
    print(payload)
    print()


def main(args=None):
    opts = parser().parse_args(args)

    if opts.log is not None:
        logging.basicConfig(level=opts.log.upper())

    xmpp = LVAlertClient(username=opts.username,
                         resource=opts.resource,
                         server=opts.server,
                         netrc=opts.netrc,
                         interactive=True)

    if not xmpp.connect(reattempt=False):
        sys.exit(1)

    try:
        if opts.action == 'listen':
            xmpp.listen(show)
            xmpp.process(block=True)
        else:
            xmpp.auto_reconnect = False
            xmpp.process(block=False)
            if opts.action == 'nodes':
                print(*xmpp.get_nodes(), sep='\n')
            elif opts.action == 'subscriptions':
                print(*xmpp.get_subscriptions(), sep='\n')
            elif opts.action == 'subscribe':
                xmpp.subscribe(*opts.node)
            elif opts.action == 'unsubscribe':
                xmpp.unsubscribe(*opts.node)
            elif opts.action == 'send':
                for openfile in opts.eventfile:
                    eventfile = openfile.read()
                    xmpp.publish(node=opts.node[0], msg=eventfile)
                    openfile.close()
    except sleekxmpp.exceptions.IqError as e:
        print('XMPP error:', e.iq['error'], file=sys.stderr)
    finally:
        xmpp.disconnect()



# Copyright (C) Patrick Brady, Brian Moe, Branson Stephens (2015)
# based on work by Jacek Konieczny <jajcus@jajcus.net> 2003-2004
#
# This file is part of lvalert
#
# lvalert is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lvalert.  If not, see <http://www.gnu.org/licenses/>.
#


"""PubSub XMPP stanza handling

Normative reference:
   ????
"""

from __future__ import print_function
__revision__="$Id$"
__docformat__="restructuredtext en"

import libxml2

## Needed because openfire has a minor bug associated with results
## being ignored in the handshake stage. There is a one line fix
## in this overloaded class.
from ligo.lvalert.lvstanzaprocessor import LVAlertStanzaProcessor


from pyxmpp.stanza import Stanza
from pyxmpp.utils import to_utf8,from_utf8
from pyxmpp.xmlextra import common_ns
from pyxmpp.objects import StanzaPayloadObject
from pyxmpp.iq import Iq
from pyxmpp.xmlextra import get_node_ns_uri
from pyxmpp.exceptions import ClientError, FatalClientError

DISCO_INFO_NS = 'http://jabber.org/protocol/disco#info'
DISCO_ITEMS_NS = 'http://jabber.org/protocol/disco#items'
JABBER_X_DATA_NS = 'jabber:x:data'

PUBSUB_NS = 'http://jabber.org/protocol/pubsub'
PUBSUB_OWNER_NS = 'http://jabber.org/protocol/pubsub#owner'
PUBSUB_NODE_CONFIG_NS = 'http://jabber.org/protocol/pubsub#node_config'

class PubSub(Iq):
    xml_element_name = "pubsub"
    xml_element_namespace = PUBSUB_NS

    def new_pubsub(self,ns_uri=PUBSUB_NS,name="pubsub"):
        ps = Stanza(name)
        ns = ps.xmlnode.newNs(ns_uri,None)
        ps.xmlnode.setNs(ns)
        return ps

    def get_pubsub(self):
        c = self.xmlnode.children
        while c:
            try:
                if c.ns():
                    return c
            except libxml2.treeError:
                pass
            c = c.next
        return None

    def get_pubsub_ns(self):
        """Get a namespace of the stanza payload.

        :return: XML namespace URI of the payload or None if there is no
            payload.
        :returntype: `str`"""
        q=self.get_query()
        if q:
            return get_node_ns_uri(q)
        else:
            return None

    def get_affiliations(self):
        ps = self.new_pubsub()
        ps.add_new_content(None,"affiliations")
        self.add_content(ps.xmlnode)

    def get_nodes(self,ns_uri="http://jabber.org/protocol/disco#items",name="query"):
        self.set_new_content(ns_uri,name)
        self.set_type("get")

    def get_nodes_result(self,cresult):
        c = cresult.xmlnode.children
        c = c.children
        print("List of nodes")
        while c:
            print("Node: %s, %s" % (c.prop("node"), c.prop("name"))) 
            c = c.next
        #print "End list of nodes"

    def create_node(self,node_name):
        ps = self.new_pubsub()
        c = ps.add_new_content(None,"create")
        c.newProp("node",node_name)
        ps.add_new_content(None,"configure")
        self.add_content(ps.xmlnode)
        self.set_type("set")

    def generic_result(self,cresult):
        print("Successfully completed operation")

    def create_timeout(self,ctimeout):
        """Process session request time out.  
        :raise FatalClientError:"""
        raise FatalClientError("Timeout while trying to create node")

    def create_error(self,ctimeout):
        """Process session request time out.  
        :raise FatalClientError:"""
        raise FatalClientError("Error while trying to create node")

    def get_node_config(self,node_name):
        ps = self.new_pubsub(ns_uri=PUBSUB_OWNER_NS)
        c = ps.add_new_content(None,"configure")
        c.newProp("node",node_name)
        self.add_content(ps.xmlnode)
        self.set_type("get")

    def delete_node(self,node_name):
        ps = self.new_pubsub(ns_uri=PUBSUB_OWNER_NS)
        c = ps.add_new_content(None,"delete")
        c.newProp("node",node_name)
        self.add_content(ps.xmlnode)
        self.set_type("set")

    def publish(self,content,node_name):
        ps = self.new_pubsub(ns_uri=PUBSUB_NS)
        c = ps.add_new_content(None,"publish")
        c.newProp("node",node_name)
        c = c.newChild(None,"item",None)
        c = c.newTextChild(None,"entry",content)
        self.add_content(ps.xmlnode)
        self.set_type("set")

    def subscriptions(self,jid):
        ps = self.new_pubsub(ns_uri=PUBSUB_NS)
        c = ps.add_new_content(None,"subscriptions")
        self.add_content(ps.xmlnode)
        self.set_type("set")

    def subscriptions_result(self,cresult):
        c = cresult.xmlnode.children.children.children
        while c:
            print("Node: %s [ %s subid=%s]" % ( c.prop("node"),\
                c.prop("subscription"), c.prop("subid") ))
            c = c.next

    def publisher(self,jid,node_name,affiliation):
        ps = self.new_pubsub(ns_uri=PUBSUB_OWNER_NS)
        c = ps.add_new_content(None,"affiliations")
        c.newProp("node",node_name)
        c = c.newChild(None,"affiliation",None)
        c.newProp("jid",jid.bare().as_utf8())
        c.newProp("affiliation",affiliation)
        self.add_content(ps.xmlnode)
        self.set_type("set")

    def affiliations(self,jid,node_name):
        ps = self.new_pubsub(ns_uri=PUBSUB_OWNER_NS)
        c = ps.add_new_content(None,"affiliations")
        c.newProp("node",node_name)
        self.add_content(ps.xmlnode)
        self.set_type("get")

    def affiliations_result(self,cresult):
        c = cresult.xmlnode.children.children.children
        while c:
            print("Affiliation: %s [ %s ]" % ( c.prop("jid"),\
                c.prop("affiliation")))
            c = c.next

    def subscribe(self,jid,node_name):
        ps = self.new_pubsub(ns_uri=PUBSUB_NS)
        c = ps.add_new_content(None,"subscribe")
        c.newProp("node",node_name)
        c.newProp("jid",jid.bare().as_utf8())
        self.add_content(ps.xmlnode)
        self.set_type("set")

    def unsubscribe(self,jid,node_name,node_subid=None):
        ps = self.new_pubsub(ns_uri=PUBSUB_NS)
        c = ps.add_new_content(None,"unsubscribe")
        c.newProp("node",node_name)
        if node_subid:
            c.newProp("subid",node_subid)
        c.newProp("jid",jid.bare().as_utf8())
        self.add_content(ps.xmlnode)
        self.set_type("set")
        self.set_from(jid)

    def subscribe_result(self,cresult):
        node_name = cresult.get_query().children.prop("node")
        print("Successfully subscribed to node %s" % node_name)

    def subscribe_timeout(self,ctimeout):
        """Process session request time out.  
        :raise FatalClientError:"""
        raise FatalClientError("Timeout while trying to subscribe to node")

    def subscribe_error(self,ctimeout):
        """Process session request time out.  
        :raise FatalClientError:"""
        raise FatalClientError("Error while trying to subscribe to node")

    def generic_timeout(self):
        """Process session request time out.  
        :raise FatalClientError:"""
        raise FatalClientError("Timeout while trying to perform action")

    def generic_error(self,iq):
        """Process session request failure.
        :Parameters:
            - `iq`: IQ error stanza received as result of the session request.
        :Types:
            - `iq`: `pyxmpp.Iq`

        :raise FatalClientError:"""
        err=iq.get_error()
        msg=err.get_message()
        raise FatalClientError("Failed to establish a session: "+msg)

class PubSubMessage(Stanza):
    """Wraper object for <message /> stanzas."""
    stanza_type="message"
    def __init__(self, xmlnode = None, from_jid = None, to_jid = None,\
                stanza_type = None, stanza_id = None, error = None,\
                error_cond = None, stream = None):
        """Initialize a `PubSubMessage` object..........
            - `error_cond`: `unicode`"""

        self.xmlnode=None
        if isinstance(xmlnode,PubSubMessage):
            pass
        elif isinstance(xmlnode,Stanza):
            raise TypeError("Couldn't make PubSubMessage from other Stanza")
        elif isinstance(xmlnode,libxml2.xmlNode):
            pass
        elif xmlnode is not None:
            raise TypeError("Couldn't make PubSubMessage from %r" % (type(xmlnode),))

        if xmlnode is None:
            xmlnode="message"

        Stanza.__init__(self, xmlnode, from_jid = from_jid, to_jid = to_jid,\
                    stanza_type = stanza_type, stanza_id = stanza_id, \
                    error = error, error_cond = error_cond, stream = stream)

    def get_event(self):
        """Get the message event

        :return: the message event or `None` if there is no event
        :returntype: `unicode`"""
        n=self.xpath_eval("ns:event")
        if n:
            return from_utf8(n[0].getContent())
        else:
            return None

# vi: sts=4 et sw=4 
